# -*- coding: utf-8 -*-
# @Time    : 2021/12/8 11:39
# @Author  : xuyiqing
# @FileName: apps.py
import logging

from django.apps import AppConfig
from django.conf import settings

from nacos_app.balanced import LoadBalanced


logger = logging.getLogger("nacos")
if not logger.handlers:
    logger = logging.getLogger("django")

if getattr(settings, "NACOS_SERVER_DISCOVERY", None) is not None:
    loadbalanced = LoadBalanced(**settings.NACOS_SERVER_DISCOVERY)


class NacosRegisterConfig(AppConfig):

    name = "nacos_app"

    def ready(self):
        if getattr(settings, "NACOS_SERVER_DISCOVERY", None) is not None:
            loadbalanced.async_handle()



# -*- coding: utf-8 -*-
# @Time    : 2022/5/6 10:33
# @Author  : xuyiqing
# @FileName: loadbalanced.py
import re
import random

import requests

from nacos_app.service import NacosServerInstance


class LoadBalanced(NacosServerInstance):
    """
    url = "http://service-name/api/test"
    """

    def __init__(self, **kwargs):
        super(LoadBalanced, self).__init__(**kwargs)

    def service_url(self, url):
        instance_names = re.findall("http[s]?://(.+?)/.+", url)
        if instance_names:
            instance_name = instance_names[0]
        else:
            raise Exception("{} url错误, 无法解析出服务名称".format(url))
        instances = self.get_instances(instance_name)
        if not instances:
            raise Exception("{} 服务无法发现".format(instance_name))
        instance = random.sample(instances, 1)[0]
        url = url.replace(instance_name, instance.host, 1)
        return url

    def request(self, method, url, params=None, body=None):
        url = self.service_url(url)
        return requests.request(method, url, params=params, json=body).json()

    def get(self, url, params=None):
        url = self.service_url(url)
        return requests.get(url, params=params).json()

    def post(self, url, body, params=None):
        url = self.service_url(url)
        return requests.post(url, json=body, params=params).json()

    def formpost(self, url, form, params=None):
        url = self.service_url(url)
        return requests.post(url, data=form, params=params).json()



# -*- coding: utf-8 -*-
# @Time    : 2022/5/6 9:46
# @Author  : xuyiqing
# @FileName: register.py
import datetime
import json
import logging
import random
import time

import requests

from nacos_app.util import get_innerip

logger = logging.getLogger("nacos")
if not logger.handlers:
    logger = logging.getLogger("django")


class NacosServerDiscoveryRegister(object):

    def __init__(self, **kwargs):
        assert kwargs.get("server_addr"), "nacos服务注册地址不可为空"
        assert kwargs.get("namespace"), "命名空间不可为空"
        assert kwargs.get("group_name"), "组名不可为空"
        assert kwargs.get("port"), "本机服务端口不可为空"
        assert kwargs.get("service_name"), "服务名称不可为空"
        assert kwargs.get("username"), "授权账号不可为空"
        assert kwargs.get("password"), "授权密码不可为空"

        self.server_addrs = kwargs["server_addr"].strip(",").split(",")
        assert self.server_addrs, "nacos服务注册地址不可为空"
        self.namespace = kwargs["namespace"]
        self.group_name = kwargs["group_name"]
        ip = get_innerip()
        if ip:
            self.ip = ip
        else:
            assert kwargs.get("ip"), "本机服务ip不可为空"
            self.ip = kwargs["ip"]
        self.port = kwargs["port"]
        self.service_name = kwargs["service_name"]
        self.ephemeral = kwargs.get("ephemeral") if isinstance(kwargs.get("ephemeral"), bool) else True
        self.username = kwargs["username"]
        self.password = kwargs["password"]
        self.access_token = ""
        self.metadata = {"beat": "true"}
        self.heart_beat = kwargs.get("heartbeat_interval") or 5

    def get_server_addr(self):
        return random.sample(self.server_addrs, 1)[0]

    def send_request(self, method, url, params=None, body=None):
        """
        发送http请求，与nacos通信
        :param method: http method
        :param url: http url
        :param params: http query 参数
        :param body: http body 参数
        :return: http response
        """
        for i in range(5):
            try:
                resp = requests.request(method, url, params=params, json=body)
            except Exception as err:
                logger.error(err)
                continue
            else:
                if resp is None:
                    logger.info("请求返回为空，-> {}次continue".format(i + 1))
                    continue
                if resp.status_code == 403:
                    self.get_access_token()
                    if "accessToken" in params:
                        params.update(accessToken=self.access_token)
                    continue
                return resp

    def get_access_token(self):
        """
        申请nacos授权token
        :return: accessToken
        """
        url = "http://{}/nacos/v1/auth/login".format(self.get_server_addr())
        params = dict(
            username=self.username,
            password=self.password
        )
        resp = self.send_request(method="POST", url=url, params=params, body=params)
        if resp.status_code == 200:
            data = json.loads(resp.text)
            logger.info("获取token成功->" + str(data))
            self.access_token = data.get("accessToken")
            return data.get("accessToken")
        else:
            logger.info("获取token失败->" + str(resp) + resp.text)

    def send_heartbeat(self):
        """
        发送心跳检索机制，第一次心跳检测等同于服务注册
        """
        url = "http://{}/nacos/v1/ns/instance/beat".format(self.get_server_addr())
        beat_data = {
            "serviceName": self.service_name,
            "ip": self.ip,
            "port": self.port,
            "weight": 1,
            "ephemeral": self.ephemeral,
            "metadata": self.metadata  # 必填
        }
        params = dict(
            serviceName=self.service_name,
            groupName=self.group_name,
            ephemeral=self.ephemeral,
            beat=json.dumps(beat_data),
            accessToken=self.access_token,
            namespaceId=self.namespace
        )
        resp = self.send_request("PUT", url, params)
        if resp.status_code == 200:
            logger.info("heartbeat success" + str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")))
        elif resp.status_code == 403:
            self.access_token = self.get_access_token()
            logger.info("heartbeat auth failed" + str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")))
        else:
            logger.info("heartbeat failed" + str(resp) + resp.text)

    def async_handle(self):
        """
        线程异步处理心跳检测，与gunicorn子进程不堵塞，需要gunicorn预加载app，添加--preload启动参数
        :return:
        """
        def run():
            while True:
                try:
                    self.send_heartbeat()
                except Exception as err:
                    logger.info("send_hearbeat 失败, {}s后重试，{}".format(self.heart_beat, err))
                time.sleep(self.heart_beat)

        self.access_token = self.get_access_token()
        import threading
        thread = threading.Thread(target=run)
        thread.setDaemon(True)
        thread.start()

    def with_nacos_sdk(self):
        import nacos
        client = nacos.NacosClient(self.get_server_addr(),
                                   namespace=self.namespace,
                                   username=self.username,
                                   password=self.password)

        def run():
            while True:
                t = client.send_heartbeat(service_name=self.service_name,
                                          ip=self.ip,
                                          port=self.port,
                                          group_name=self.group_name,
                                          metadata=json.dumps({"beat": "true"}))
        from concurrent.futures import ThreadPoolExecutor
        pool = ThreadPoolExecutor(max_workers=1)
        pool.submit(run)



# -*- coding: utf-8 -*-
# @Time    : 2022/5/6 9:39
# @Author  : xuyiqing
# @FileName: service.py
import logging

from nacos_app.obj import Instance
from nacos_app.register import NacosServerDiscoveryRegister


logger = logging.getLogger("nacos")
if not logger.handlers:
    logger = logging.getLogger("django")


class NacosServerInstance(NacosServerDiscoveryRegister):

    def __init__(self, **kwargs):
        super(NacosServerInstance, self).__init__(**kwargs)

    def get_instances(self, service_name: str):
        url = "http://{}/nacos/v1/ns/instance/list".format(self.get_server_addr())
        params = {
            "serviceName": service_name,
            "groupName": self.group_name,
            "namespaceId": self.namespace,
            "healthyOnly": True,
            "accessToken": self.access_token
        }
        resp = self.send_request("GET", url, params=params).json()
        hosts = resp["hosts"]
        instances = []
        for host in hosts:
            instance = Instance(**host)
            instances.append(instance)
        return instances



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

current_dir = os.path.dirname(os.path.realpath(__file__))

packages = [
    'nacos_app'
]

long_description = """
1、适用于Django服务程序注册微服务实例至nacos服务中心;
2、实现了服务与注册中心的登录授权，服务注册，心跳检测，服务调用负载均衡;
3、cloud模块作为一个django app注册，在INSTALLED_APPS加入'nacos_app.apps.NacosRegisterConfig'
4、django程序需要在settings环境中配置服务注册的信息NACOS_SERVER_DISCOVERY
5、gunicorn多个worker模式下启动需要添加--preload参数，由管理进程预加载非函数式编程中的代码块，从而避免多个worker同时加载register
   gunicorn backend.wsgi -w 8 -b 0.0.0.0:port -t 600 --preload
NACOS_SERVER_DISCOVERY = {
    "server_addr": "",                    # nacos服务中心地址
    "namespace": "",                      # 命名空间
    "group_name": "",                     # 分组
    "ip": "",                             # 本机ip
    "port": "",                           # 本机服务端口
    "service_name": "",                   # 本机服务名称
    "ephemeral": True,                    # 是否临时实例，true为临时实例，临时实例用于服务临时扩展，过后丢弃
    "username": "",                       # 拥有对应命名空间权限的账户
    "password": "",                       # 密码
    "heartbeat_interval": 6               # 心跳检测间隔，单位秒
}
"""

setup(
    name='nacos_app',
    version='1.1.2',
    description='djagno-nacos-app',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email="1593134926@qq.com",
    author="徐益庆",
    packages=packages,
    package_dir={'nacos_app': 'nacos_app'},
    include_package_data=True,
    python_requires='>=2.6, !=3.0.*, !=3.1.*, !=3.2.*, <4',
    zip_safe=False,
)



# -*- coding: utf-8 -*-
# @Time    : 2022/5/6 9:51
# @Author  : xuyiqing
# @FileName: util.py
import logging
import socket

import requests

logger = logging.getLogger("nacos")
if not logger.handlers:
    logger = logging.getLogger("django")


def get_innerip():
    s, ip = None, None
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    finally:
        if s is not None:
            s.close()
    logger.info("获取本机局域网ip->{}".format(ip))
    return ip



# -*- coding: utf-8 -*-
# @Time    : 2022/5/6 10:26
# @Author  : xuyiqing
# @FileName: obj.py

class Instance(object):

    def __init__(self, **kwargs):
        self.data = kwargs
        self.ip = kwargs.get("ip")
        self.port = kwargs.get("port")
        self.service_name = kwargs.get("serviceName")
        self.instance_id = kwargs.get("instanceId")
        self.healthy = kwargs.get("healthy")
        self.host = "{}:{}".format(self.ip, self.port)

    def __repr__(self):
        return "Instance--{}".format(self.data)
